// -------------------------------------------------------------------------------------------------
// A simple parallel implementation for Roger Alsing's EvoLisa problem in less than 400 effective
// lines of C++ code.
// Mostly for educational purposes (clear, readable code is preferred).
// It is used in the "Algorithms and Complexity" course at Budapest University of Technology
// and Economics.
//
// Károly Zsolnai, zsolnai@iit.bme.hu
// http://cg.iit.bme.hu/~zsolnai/
//
// Roger's site:
// http://rogeralsing.com/2008/12/16/evolisa-optimizations-and-improved-quality/
//
// See c.bat for compilation under Windows with MinGW or TDM GCC.
//
// Compilation under Linux:
// gcc *.cpp -O3 -ffast-math -lglut -lGL -lGLU -fopenmp
// for more optimizations: -mfpmath=sse -msse -msse2 -msse3
// -------------------------------------------------------------------------------------------------
// Note: the glReadPixels() call substantially hinders performance, especially with shared-memory
// VGA cards on notebooks.
// Speed above ~40 gens/sec may be considered acceptable, I have around 100 gens/sec on an
// average 4-core desktop computer as of 2012. If you find it too slow, just try adjusting the 
// POPSIZE and POLYCOUNT macros and/or installing the newest driver for your VGA may also be useful.
// -------------------------------------------------------------------------------------------------
// Gabriele Greco had also made a blazing fast SDL implementation:
// source: http://www.ggsoft.org/archives/genetic.zip
// binary: http://www.ggsoft.org/archives/genetic_test.zip (it may require SDL.dll for Windows, just
// google it and download it from anywhere)
// -------------------------------------------------------------------------------------------------

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include <cmath>
#include <algorithm>
#include <GL/glut.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <omp.h>

#include "RgbImage.h"

int width = 800;
int height = 800;

char title[128];

int mousex, mousey;
int old_fps;

int curr_time, d_time, fps, gens, threads = 0;

bool keysPressed[256];
unsigned char* pic;
//const char* filename = "red-green-200.bmp"; // it is a simple test case
const char* filename = "mona-200.bmp"; // the real challenge 
RgbImage image(filename);

#if defined(__linux) || defined(__linux__)
	unsigned int seed = time(NULL);
	#define RND (2.0*(double)rand_r(&seed)/RAND_MAX-1.0)
	#define RND2 ((double)rand_r(&seed)/RAND_MAX) // uniform rnd
#endif

#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__)
	#define RND (2.0*(double)rand()/RAND_MAX-1.0)
	#define RND2 ((double)rand()/RAND_MAX) // uniform rnd
	#define round(x) floor(x+0.5)

typedef BOOL (APIENTRY *PFNWGLSWAPINTERVALFARPROC)( int );
PFNWGLSWAPINTERVALFARPROC wglSwapIntervalEXT = 0;

void setvsync(int interval=1) {
  const char *extensions = glGetString( GL_EXTENSIONS );
  if( strstr( extensions, "WGL_EXT_swap_control" ) == 0 ) return;
  else {
    wglSwapIntervalEXT = (PFNWGLSWAPINTERVALFARPROC)wglGetProcAddress( "wglSwapIntervalEXT" );
    if( wglSwapIntervalEXT ) wglSwapIntervalEXT(interval);
  }
}
#endif

// POPSIZE - the number of chromosomes in the population
// POLYCOUNT - the number of triangles for every chromosome
// VERTICES - 3 to draw triangles, modify draw() for quads before changing this
// DIMS - 2 dimensional drawing
#define POPSIZE 30
#define POLYCOUNT 150
#define VERTICES 3
#define DIMS 2

const int disc = (int)(ceil(POPSIZE*0.75)); // chromosomes discarded via elitism
const float crp = 0.95; // crossover probability (therefore mutation probability is 1-crp)

struct chromo {
	float vertices[POLYCOUNT][VERTICES][DIMS];
	float colors[POLYCOUNT][4];
	uint_fast64_t fitness;
	unsigned char* screen;

	chromo() { screen = (unsigned char*)malloc(sizeof(unsigned char)*image.w*image.h*3); }
	~chromo() {	}
	void draw() {
		glBegin(GL_TRIANGLES);
		for(int i=0;i<POLYCOUNT;i++) {
			glColor4f(colors[i][0], colors[i][1],colors[i][2], colors[i][3]); // RGBA
			glVertex2f(vertices[i][0][0],vertices[i][0][1]);
			glVertex2f(vertices[i][1][0],vertices[i][1][1]);
			glVertex2f(vertices[i][2][0],vertices[i][2][1]);
		}
		glEnd();
	}
	// The fitness is a simple squared distance between the target and synthesized image
	inline uint_fast64_t calc_fitness() {
		// uint_fast64_t is used because it may be quite a huge value 
		uint_fast64_t fit = 0;
	    glClear(GL_COLOR_BUFFER_BIT);
	    draw();
		// this is the bottleneck
		glReadPixels(0,0,image.w,image.h,GL_RGB,GL_UNSIGNED_BYTE,screen);
	    glClear(GL_COLOR_BUFFER_BIT);

//		Partition into equal parts and assign them to threads
//		#pragma omp parallel for schedule(static,(int)((image.w*image.h*3)/threads)) reduction(+:fit)
//		guided approach with a chunk size of 1024
		#pragma omp parallel for schedule(guided,1024) reduction(+:fit)
		for(int i=0;i<image.w*image.h*3;i+=3) {
			fit += ((screen[i] - image.ImagePtr[i])*(screen[i] - image.ImagePtr[i])) +
    				((screen[i+1] - image.ImagePtr[i+1])*(screen[i+1] - image.ImagePtr[i+1])) +
					((screen[i+2] - image.ImagePtr[i+2])*(screen[i+2] - image.ImagePtr[i+2]));
		}
		return fit;
	}
	// Mutation strategy #1: random perturbations to the coordinates and colors of vertices
	void mutate1(float perturbation) {
		for(int j=0;j<POLYCOUNT;j++) {
			for(int k=0;k<VERTICES;k++) {
				for(int l=0;l<DIMS;l++) {
					if(RND2<0.25f) vertices[j][k][l] += RND/perturbation;
					if( vertices[j][k][l] < .0f || vertices[j][k][l] > 1.f) vertices[j][k][l] = RND2;
				}
			}
		if(RND2<0.5f) {
			colors[j][0] += 10*RND/perturbation;
			colors[j][1] += 10*RND/perturbation;
			colors[j][2] += 10*RND/perturbation;
		}
		if( colors[j][0] < .0f || colors[j][0] > 1.f) colors[j][0] = RND2;
		if( colors[j][1] < .0f || colors[j][1] > 1.f) colors[j][1] = RND2;
		if( colors[j][2] < .0f || colors[j][2] > 1.f) colors[j][2] = RND2;
		}
	}
	// Mutation strategy #2: completely new set of vertices and color
	void mutate2() {
		for(int j=0;j<POLYCOUNT;j++) {
			for(int k=0;k<VERTICES;k++) {
				for(int l=0;l<DIMS;l++) {
					if(RND2>0.5f) vertices[j][k][l] = RND2;
				}
			}
			if(RND2>0.5f) {
				colors[j][0] = RND2;
				colors[j][1] = RND2; 
				colors[j][2] = RND2;
			}
		}
	}
};

// the chromosome population
chromo pop[POPSIZE];

// initialize the chromosomes in the population with random values
void init_pop(chromo* pop) {
	for(int i=0;i<POPSIZE;i++) {
		for(int j=0;j<POLYCOUNT;j++) {
			for(int k=0;k<VERTICES;k++) {
				for(int l=0;l<DIMS;l++) {
					pop[i].vertices[j][k][l] = RND2;
				}
			}
		pop[i].colors[j][0] = RND2;
		pop[i].colors[j][1] = RND2;
		pop[i].colors[j][2] = RND2;
		pop[i].colors[j][3] = 0.15; // this is the alpha value
		}
	}
}

bool cmp(const chromo& a, const chromo& b) {
	return a.fitness < b.fitness;
}

void init() {
	if( image.GetErrorCode() == 0) printf("\nLoaded file \"%s\", width: %ld, height: %ld.\n",filename,image.w,image.h);
	srand(time(NULL));

	// Show the number of available CPU cores
	#pragma omp parallel
	{ threads = omp_get_num_threads();}
	printf("Launching %d threads.",threads);

	// Population initialization pass
	for(int i=0;i<POPSIZE;i++) { pop[i] = chromo(); }
	init_pop(pop);
	for(int i=0;i<POPSIZE;i++) {
		pop[i].fitness = pop[i].calc_fitness();
	}
	std::sort(pop, pop+POPSIZE, cmp);
}

// One point crossover scheme:
// inherit the first parent's vertices and colors until the crossover point and
// the second's thereafter
void crossover1p(const chromo c1, const chromo c2, chromo& c3) {
	int cp = round(RND2*POLYCOUNT); // crossover point
	for(int i=0;i<POLYCOUNT;i++) {
		for(int j=0;j<VERTICES;j++) {
			for(int k=0;k<DIMS;k++) {
				if(i<cp) c3.vertices[i][j][k] = c1.vertices[i][j][k]; // make a for later on
				else c3.vertices[i][j][k] = c2.vertices[i][j][k];
			}
		}
	}
	for(int i=0;i<POLYCOUNT;i++) {
		if(i<cp) { 
			c3.colors[i][0] = c1.colors[i][0];
			c3.colors[i][1] = c1.colors[i][1];
			c3.colors[i][2] = c1.colors[i][2];
		}
		else {
		 	c3.colors[i][0] = c2.colors[i][0];
		 	c3.colors[i][1] = c2.colors[i][1];
		 	c3.colors[i][2] = c2.colors[i][2];
		}
	}
}

void crossoverrand(const chromo c1, const chromo c2, chromo& c3) {
	for(int i=0;i<POLYCOUNT;i++) {
		for(int j=0;j<VERTICES;j++) {
			if(RND2<0.5) {
				for(int k=0;k<DIMS;k++) {
					c3.vertices[i][j][k] = c1.vertices[i][j][k];
					c3.colors[i][0] = c1.colors[i][0];
					c3.colors[i][1] = c1.colors[i][1];
					c3.colors[i][2] = c1.colors[i][2];
				}
			}
			else {
				for(int k=0;k<DIMS;k++) {
					c3.vertices[i][j][k] = c2.vertices[i][j][k];
					c3.colors[i][0] = c2.colors[i][0];
					c3.colors[i][1] = c2.colors[i][1];
					c3.colors[i][2] = c2.colors[i][2];
				}
			}
		}
	}
}

void drawText(char *str, void *font, GLclampf r, GLclampf g, GLclampf b,
              GLfloat x, GLfloat y) {  
	char *ch;  
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0, 1.0, 0.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glPushAttrib(GL_COLOR_BUFFER_BIT);       
	glColor4f(r, g, b, 1);
	glRasterPos3f(x, y, 0.0);
	for(ch= str; *ch; ch++) {
		glutBitmapCharacter(font, (int)*ch);
	}
	glPopAttrib();
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix(); 
}

void display() {
	glViewport( 0, 0, width, height );
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluOrtho2D(0.0, 1.0, 0.0, 1.0);
    glClear(GL_COLOR_BUFFER_BIT);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// always draw the best chromosome
	pop[0].draw();
	char buf[128], buf2[128], buf3[128], buf4[128], buf5[128];
	sprintf(buf, "Generations: %d", gens);
	sprintf(buf2, "Fitness: %lld", pop[0].fitness/30000);
	sprintf(buf3, "Time: %2.3fs", (float)glutGet(GLUT_ELAPSED_TIME)/1000);
	sprintf(buf4, "Gens/sec: %2.1f", gens/((float)glutGet(GLUT_ELAPSED_TIME)/1000));
	sprintf(buf5, "Threads: %d", threads);
	float start = 0.93f;
	float diff = .065f;
	drawText(buf,GLUT_BITMAP_HELVETICA_12, 0.9,0.9,0.9, 0.03,start);
	drawText(buf2,GLUT_BITMAP_HELVETICA_12, 0.9,0.9,0.9, 0.03,start-diff);
	drawText(buf3,GLUT_BITMAP_HELVETICA_12, 0.9,0.9,0.9, 0.03,start-2*diff);
	drawText(buf4,GLUT_BITMAP_HELVETICA_12, 0.9,0.9,0.9, 0.03,start-3*diff);
	drawText(buf5,GLUT_BITMAP_HELVETICA_12, 0.9,0.9,0.9, 0.03,start-4*diff);

    glutSwapBuffers();
}

void idle() {
	gens++;
	d_time = glutGet(GLUT_ELAPSED_TIME) - curr_time;
	if(d_time) fps = (1000/d_time);
	curr_time = glutGet(GLUT_ELAPSED_TIME);

	sprintf(title, "Genetic, %d FPS",fps);
	glutSetWindowTitle(title);

	// Compute new population
	std::sort(pop, pop+POPSIZE, cmp);
	for(int i=0;i<POPSIZE;i++) {
		if(i>POPSIZE-disc) {
			if(RND2<crp) { // crossover section
				int ind1 = ((int)round(RND2*POPSIZE))%POPSIZE;
				int ind2 = ((int)round(RND2*POPSIZE))%POPSIZE;
				if(RND2<0.5) crossover1p(pop[ind1],pop[ind2],pop[i]);
				else crossoverrand(pop[ind1],pop[ind2],pop[i]);
			}
			else { // mutation
				if(RND2<0.95) pop[i].mutate1(500*RND); // tries multiple little changes to the chromosomes
				else pop[i].mutate2(); // reinitializes the chromosome with random values
			}
		}
	}
	// fitness recalculation only for the newly created chromosomes
	for(int i=0;i<POPSIZE;i++) {
		if(i>POPSIZE-disc) {
			pop[i].fitness = pop[i].calc_fitness();
		}
	}

  glutPostRedisplay();
}

void keydown(unsigned char key, int x, int y) {
  keysPressed[key] = true;
}

void keyup(unsigned char key, int x, int y) {
  keysPressed[key] = false;
  switch(key) {
	  case (13):
		exit(0);
		break;
	  case (27):
		exit(0);
		break;
  }
}

void mouse(int button, int state, int x, int y) {
  if(button == GLUT_LEFT_BUTTON)
    if(state == GLUT_DOWN) {
      mousex = x;
      mousey = y;
    }
}

void motion(int x, int y) {
}

void reshape(int w, int h) {
	width = w;
	height = h;
}

int main(int argc, char** argv) {
  glutInit(&argc, argv);
  glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB);

  width = image.w;
  height = image.h; 
  glutInitWindowSize(width, height);
  glutInitWindowPosition(300, 200);
  glutCreateWindow("Genetic");

#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__)
  setvsync(0);
#endif
  
  init();

  glutDisplayFunc(display);
  glutIdleFunc(idle);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keydown);
  glutKeyboardUpFunc(keyup);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);

  glutMainLoop();	

  return 0; 
}
